from __future__ import division

import collections
import datetime
from math import log10

import numpy as np
import pandas as pd


DOCUMENT_DOES_NOT_EXIST = 'The specified document does not exist'
TERM_DOES_NOT_EXIST = 'The specified term does not exist'


class HashedIndex(object):
    """
    InvertedIndex structure in the form of a hash list implementation.
    """

    def __init__(self, initial_terms=None):
        """
        Construct a new HashedIndex. An optional list of initial terms
        may be passed which will be automatically added to the new HashedIndex.
        """
        self._documents = collections.Counter()
        self._terms = {}
        self._freeze = False
        if initial_terms is not None:
            for term in initial_terms:
                self._terms[term] = {}

    def __getitem__(self, term):
        return self._terms[term]

    def __contains__(self, term):
        return term in self._terms

    def __repr__(self):
        return '<HashedIndex: {} terms, {} documents>'.format(
            len(self._terms), len(self._documents)
        )

    def __eq__(self, other):
        return self._terms == other._terms and self._documents == other._documents

    def clear(self):
        """
        Resets the HashedIndex to a clean state without any terms or documents.
        """
        self._terms = {}
        self._documents = collections.Counter()

    def freeze(self):
        """
        Freezes the HashedIndex, preventing any new terms from being added
        when calling add_term_occurrence.
        """
        self._freeze = True

    def unfreeze(self):
        """
        Unfreezes (thaws) the HashedIndex, allowing new terms to be added
        when calling add_term_occurrence.
        """
        self._freeze = False

    def add_term_occurrence(self, term, document):
        """
        Adds an occurrence of the term in the specified document.
        """
        if document not in self._documents:
            self._documents[document] = 0

        if term not in self._terms:
            if self._freeze:
                return
            else:
                self._terms[term] = collections.Counter()

        if document not in self._terms[term]:
            self._terms[term][document] = 0

        self._documents[document] += 1
        self._terms[term][document] += 1

    def get_total_term_frequency(self, term):
        """
        Gets the frequency of the specified term in the entire corpus
        added to the HashedIndex.
        """
        if term not in self._terms:
            raise IndexError(TERM_DOES_NOT_EXIST)

        return sum(self._terms[term].values())

    def get_term_frequency(self, term, document):
        """
        Returns the frequency of the term specified in the document.
        """
        if document not in self._documents:
            raise IndexError(DOCUMENT_DOES_NOT_EXIST)

        if term not in self._terms:
            raise IndexError(TERM_DOES_NOT_EXIST)

        if document not in self._terms[term]:
            return 0

        return self._terms[term][document]

    def get_document_frequency(self, term):
        """
        Returns the number of documents the specified term appears in.
        """
        if term not in self._terms:
            raise IndexError(TERM_DOES_NOT_EXIST)
        else:
            return len(self._terms[term])

    def get_document_length(self, document):
        """
        Returns the number of terms found within the specified document.
        """
        if document in self._documents:
            return self._documents[document]
        else:
            raise IndexError(DOCUMENT_DOES_NOT_EXIST)

    def terms(self):
        return self._terms.keys()

    def documents(self):
        return list(self._documents)

    def items(self):
        return self._terms

    def get_tfidf(self, term, document):
        """
        Returns the Term-Frequency Inverse-Document-Frequency value for the given
        term in the specified document.
        """
        tf = self.get_term_frequency(term, document)

        # Speeds up performance by avoiding extra calculations
        if tf != 0.0:
            # Add 1 to document frequency to prevent divide by 0
            df = 1 + self.get_document_frequency(term)
            n = 1 + len(self._documents)

            return tf * log10(n / df)
        else:
            return 0.0

    def get_total_tfidf(self, term):
        result = 0
        for document in self._documents:
            result += self.get_tfidf(term, document)
        return result

    def generate_document_vector(self, doc, mode='tfidf'):
        """
        Returns a representation of the specified document as a feature vector
        weighted according the mode specified (by default tf-dif). The result
        will be returned in the form of a numpy ndarray.
        Available modes: tfidif, count, tf
        """
        result = np.zeros(len(self._terms))
        for i, term in enumerate(self._terms):
            if mode == 'tfidf':
                result[i] = self.get_tfidf(term, doc)
            elif mode == 'count':
                result[i] = self.get_term_frequency(term, doc)
            elif mode == 'tf':
                result[i] = self.get_term_frequency(term, doc) / self.get_document_length(doc)
            else:
                raise ValueError('Unexpected mode: %s', mode)

        return result

    def generate_feature_matrix(self, mode='tfidf'):
        """
        Returns a feature numpy matrix representing the terms and
        documents in this Inverted Index using the tf-idf weighting
        scheme by default. The term counts in each document can
        alternatively be used by specifying scheme='count'

        The size of the matrix is equal to m x n where m is
        the number of documents and n is the number of terms.
        """
        result = []

        for i, doc in enumerate(self._documents):
            result.append(self.generate_document_vector(doc, mode))

        return np.asmatrix(result)

    def generate_dataframe(self, mode='tfidf'):
        return pd.DataFrame(
            data=self.generate_feature_matrix(mode),
            columns=self.terms(),
            index=self.documents(),
        )

    def prune(self, min_frequency=2, max_percentile=1.0):
        """
        Prunes the HashedIndex so that terms under the specified min_frequency
        will be removed from the index. This method is useful for performing
        feature selection based on document frequency within the entire corpus.
        """
        n = float(len(self._documents))

        garbage = []
        for term in self.terms():
            freq = self.get_document_frequency(term)

            if freq < min_frequency:
                garbage.append(term)

            if freq / n > max_percentile:
                garbage.append(term)

        for term in garbage:
            del(self._terms[term])

    def to_dict(self):
        return {
            'documents': self._documents,
            'terms': self._terms,
        }

    def from_dict(self, data):
        self._documents = data['documents']
        self._terms = data['terms']

    @staticmethod
    def merge(index_list):
        result = HashedIndex()

        for index in index_list:
            first_index = result
            second_index = index

            assert isinstance(second_index, HashedIndex)

            for term in second_index.terms():
                if term in first_index._terms and term in second_index._terms:
                    result._terms[term] = first_index._terms[term] + second_index._terms[term]
                elif term in second_index._terms:
                    result._terms[term] = second_index._terms[term]
                else:
                    raise ValueError("I dont know how the hell you managed to get here")

            result._documents = first_index._documents + second_index._documents

        return result



from __future__ import division

import unittest

from hashedindex import hashedindex


def unordered_list_cmp(list1, list2):
    # Check lengths first for slight improvement in performance
    return len(list1) == len(list2) and sorted(list1) == sorted(list2)


class HashedIndexTest(unittest.TestCase):
    # Note that generate_document_vector and generate_feature_matrix tests
    # are considered interrelated and test cases are therefore not repeated
    # between them where possible.

    def setUp(self):
        self.index = hashedindex.HashedIndex()

        for i in xrange(3):
            self.index.add_term_occurrence('word', 'document1.txt')

        for i in xrange(5):
            self.index.add_term_occurrence('malta', 'document1.txt')

        for i in xrange(4):
            self.index.add_term_occurrence('phone', 'document2.txt')

        for i in xrange(2):
            self.index.add_term_occurrence('word', 'document2.txt')

    def test_hashedindex_constructor_with_terms(self):
        index2 = hashedindex.HashedIndex(self.index.terms())

        # Terms between the two indexes should be equal
        assert index2.terms() == self.index.terms()

        # No documents should be found
        assert index2.documents() == []

        # All terms should have no referenced documents
        for term in index2.terms():
            assert index2[term] == {}

        index2.add_term_occurrence('phone', 'mydoc.doc')
        assert index2.get_term_frequency('phone', 'mydoc.doc') == 1

    def test_case_sensitive_documents(self):
        self.index.add_term_occurrence('word', 'Document2.txt')

        assert self.index.get_term_frequency('word', 'document2.txt') == 2
        assert self.index.get_term_frequency('word', 'Document2.txt') == 1

        assert unordered_list_cmp(self.index.documents(), ['document1.txt', 'document2.txt', 'Document2.txt'])

    def test_getitem(self):
        assert unordered_list_cmp(self.index['word'].keys(), ['document1.txt', 'document2.txt'])
        assert unordered_list_cmp(self.index['malta'].keys(), ['document1.txt'])
        assert unordered_list_cmp(self.index['phone'].keys(), ['document2.txt'])

    def test_getitem_raises_keyerror(self):
        # Trying to get a term that does not exist should raise a key error
        self.assertRaises(KeyError, self.index.__getitem__, 'doesnotexist')

        # Case Insensitive check
        self.assertRaises(KeyError, self.index.__getitem__, 'wORd')

    def test_contains(self):
        assert 'word' in self.index
        assert 'malta' in self.index
        assert 'phone' in self.index

        # Case Insensitive Check
        assert 'WoRd' not in self.index

        # Non-Existent check
        assert 'doesnotexist' not in self.index

    def test_clear(self):
        self.index.clear()

        assert self.index.terms() == []
        assert self.index.documents() == []

    def test_freeze_unfreeze(self):
        self.index.freeze()
        self.index.add_term_occurrence('myword', 'document2.txt')

        # New words should be not added
        assert 'myword' not in self.index

        # Adding words that exist should work though
        assert self.index.get_term_frequency('word', 'document1.txt') == 3

        # Ensure documents are still added even if its term is not
        self.index.add_term_occurrence('idonotexist', 'document20.txt')
        assert 'document20.txt' in self.index.documents()

        self.index.add_term_occurrence('phone', 'document9.txt')
        self.index.add_term_occurrence('word', 'document1.txt')

        assert self.index.get_term_frequency('word', 'document1.txt') == 4
        assert self.index.get_term_frequency('phone', 'document9.txt') == 1

        # Terms should be add-able
        self.index.unfreeze()
        self.index.add_term_occurrence('myword', 'document2.txt')
        assert 'myword' in self.index

    def test_get_total_term_frequency(self):
        assert self.index.get_total_term_frequency('word') == 5
        assert self.index.get_total_term_frequency('malta') == 5
        assert self.index.get_total_term_frequency('phone') == 4

    def test_get_total_term_frequency_exceptions(self):
        self.assertRaises(IndexError, self.index.get_total_term_frequency, 'doesnotexist')

    def test_get_total_term_frequency_case(self):
        self.assertRaises(IndexError, self.index.get_total_term_frequency, 'WORD')
        self.assertRaises(IndexError, self.index.get_total_term_frequency, 'Malta')
        self.assertRaises(IndexError, self.index.get_total_term_frequency, 'phonE')

    def test_get_term_frequency(self):
        # Check Existing cases
        assert self.index.get_term_frequency('word', 'document1.txt') == 3
        assert self.index.get_term_frequency('malta', 'document1.txt') == 5
        assert self.index.get_term_frequency('phone', 'document2.txt') == 4
        assert self.index.get_term_frequency('word', 'document2.txt') == 2

        # Check non existing cases
        assert self.index.get_term_frequency('malta', 'document2.txt') == 0
        assert self.index.get_term_frequency('phone', 'document1.txt') == 0

    def test_get_term_frequency_exceptions(self):
        self.assertRaises(IndexError, self.index.get_term_frequency, 'doesnotexist', 'document1.txt')
        self.assertRaises(IndexError, self.index.get_term_frequency, 'malta', 'deoesnotexist.txt')

    def test_get_document_frequency(self):
        assert self.index.get_document_frequency('word') == 2
        assert self.index.get_document_frequency('malta') == 1
        assert self.index.get_document_frequency('phone') == 1

    def test_get_document_frequency_exceptions(self):
        self.assertRaises(IndexError, self.index.get_document_frequency, 'doesnotexist')

    def test_get_document_length(self):
        assert self.index.get_document_length('document1.txt') == 8
        assert self.index.get_document_length('document2.txt') == 6

    def test_get_document_length_exceptions(self):
        self.assertRaises(IndexError, self.index.get_document_length, 'doesnotexist.txt')

    def test_get_terms(self):
        assert unordered_list_cmp(self.index.terms(), ['word', 'malta', 'phone'])

        self.index.add_term_occurrence('test', 'document3.txt')
        assert unordered_list_cmp(self.index.terms(), ['word', 'malta', 'phone', 'test'])

        assert 'doesnotexist' not in self.index.terms()

    def test_get_items(self):
        assert self.index.items() == {
            'word': {'document1.txt': 3, 'document2.txt': 2},
            'malta': {'document1.txt': 5},
            'phone': {'document2.txt': 4}
        }

    def test_get_documents(self):
        assert unordered_list_cmp(self.index.documents(), ['document1.txt', 'document2.txt'])

        self.index.add_term_occurrence('test', 'document3.txt')
        assert unordered_list_cmp(self.index.documents(), ['document1.txt', 'document2.txt', 'document3.txt'])

        assert 'doesnotexist.txt' not in self.index.documents()

    def test_get_tfidf_relation(self):
        # Test Inverse Document Frequency
        self.assertLess(
            self.index.get_tfidf('word', 'document1.txt'),
            self.index.get_tfidf('malta', 'document1.txt')
        )

    def test_get_tfidf_non_negative(self):
        matrix = self.index.generate_feature_matrix(mode='tfidf')
        assert (matrix >= 0).all()

    def test_get_tfidf_empty_document(self):
        assert self.index.get_tfidf('malta', 'document2.txt') == 0

    def test_get_tfidf_empty_term(self):
        assert self.index.get_tfidf('phone', 'document1.txt') == 0

    def test_generate_document_vector_default(self):
        assert (self.index.generate_document_vector('document1.txt') ==
                self.index.generate_document_vector('document1.txt', mode='tfidf')).all()

    def test_generate_feature_matrix_default(self):
        assert (self.index.generate_feature_matrix() == self.index.generate_feature_matrix(mode='tfidf')).all()

    def test_generate_feature_matrix_tfidf(self):
        features = self.index.terms()
        instances = self.index.documents()

        matrix = self.index.generate_feature_matrix(mode='tfidf')

        assert matrix[instances.index('document1.txt'), features.index('malta')] \
            == self.index.get_tfidf('malta', 'document1.txt')

        assert matrix[instances.index('document2.txt'), features.index('word')] \
            == self.index.get_tfidf('word', 'document2.txt')

        assert matrix[instances.index('document2.txt'), features.index('phone')] \
            == self.index.get_tfidf('phone', 'document2.txt')

        assert matrix[instances.index('document1.txt'), features.index('word')] \
            == self.index.get_tfidf('word', 'document1.txt')

        # Zero Cases
        assert matrix[instances.index('document2.txt'), features.index('malta')] == 0
        assert matrix[instances.index('document1.txt'), features.index('phone')] == 0

    def test_generate_document_vector_count(self):
        features = self.index.terms()
        vector = self.index.generate_document_vector('document1.txt', mode='count')

        # Correct vector shape
        assert vector.shape == (len(features),)

        assert vector[features.index('malta')] == 5.0
        assert vector[features.index('word')] == 3.0
        assert vector[features.index('phone')] == 0.0

    def test_generate_feature_matrix_count(self):
        # Extract the feature and document indices
        features = self.index.terms()
        instances = self.index.documents()

        matrix = self.index.generate_feature_matrix(mode='count')

        # Correct matrix dimensions
        assert matrix.shape == (2, 3)

        # Ensure this method of addressing data works
        assert matrix[instances.index('document1.txt'), features.index('malta')] == 5
        assert matrix[instances.index('document2.txt'), features.index('word')] == 2
        assert matrix[instances.index('document1.txt'), features.index('word')] == 3
        assert matrix[instances.index('document2.txt'), features.index('phone')] == 4

        # Zero cases
        assert matrix[instances.index('document2.txt'), features.index('malta')] == 0
        assert matrix[instances.index('document1.txt'), features.index('phone')] == 0

    def test_generate_feature_matrix_tf(self):
        features = self.index.terms()
        instances = self.index.documents()

        matrix = self.index.generate_feature_matrix(mode='tf')

        assert matrix[instances.index('document1.txt'), features.index('word')] == 3 / 8
        assert matrix[instances.index('document2.txt'), features.index('phone')] == 4 / 6
        assert matrix[instances.index('document1.txt'), features.index('malta')] == 5 / 8

        assert matrix[instances.index('document2.txt'), features.index('malta')] == 0
        assert matrix[instances.index('document2.txt'), features.index('word')] == 2 / 6

    def test_generate_feature_matrix_invalid(self):
        self.assertRaises(ValueError, self.index.generate_feature_matrix, mode='invalid')
        self.assertRaises(ValueError, self.index.generate_feature_matrix, mode=None)

    def test_prune(self):
        self.index = hashedindex.HashedIndex()  # Fresh Index

        for i in xrange(100):
            self.index.add_term_occurrence('word', 'document{}.txt'.format(i))

        for i in xrange(20):
            self.index.add_term_occurrence('text', 'document{}.txt'.format(i))

        self.index.add_term_occurrence('lonely', 'document2.txt')

        self.index.prune(min_frequency=2)
        assert 'word' in self.index.terms()
        assert 'text' in self.index.terms()
        assert 'lonely' not in self.index.terms()

        self.index.prune(min_frequency=25)
        assert 'word' in self.index.terms()
        assert 'text' not in self.index.terms()

    def test_merge_index(self):
        first_index = hashedindex.HashedIndex()
        first_index.add_term_occurrence('foo', 'document2.txt')
        first_index.add_term_occurrence('foo', 'document1.txt')

        second_index = hashedindex.HashedIndex()
        second_index.add_term_occurrence('foo', 'document1.txt')
        second_index.add_term_occurrence('bar', 'document9.txt')

        merged_index = hashedindex.HashedIndex.merge([
            first_index,
            second_index,
        ])

        assert unordered_list_cmp(
            merged_index.terms(), 
            ['foo', 'bar']
        )
        assert unordered_list_cmp(
            merged_index.documents(), 
            ['document1.txt', 'document2.txt', 'document9.txt']
        )

        assert merged_index._terms['foo'] == {'document1.txt': 2, 'document2.txt': 1}
        assert merged_index._terms['bar'] == {'document9.txt': 1}

        assert merged_index._documents['document1.txt'] == 2
        assert merged_index._documents['document2.txt'] == 1
        assert merged_index._documents['document9.txt'] == 1
